/***************************************************************************
 *
 *
 *
 *
 *
 */



#ifdef __cplusplus
extern "C" {
#endif



#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/inotify.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>

//#define WITH_MSG

#ifdef WITH_MSG
#include <sys/ipc.h>
#include <sys/msg.h>

#define MSGSZ     128

typedef struct msgbuf {
    long    mtype;
    char    mtext[MSGSZ];
} message_buf;

#define KEY1 1563
#define KEY2 2563

int msqid1, msqid2;
message_buf sbuf,rbuf;
size_t buf_length;
#endif
/***************************************************************************
 *
 *
 *
 */

/* request cycles */
#define MIN_REQUESTS	1
#define	MAX_REQUESTS	10000000
#define	DEF_REQUESTS	100000

/* shared memory data */
#define	SHMEM_FILE_PATH	"shared.data"
#define	SHMEM_FILE_SIZE	0x1000



/***************************************************************************
 *
 *
 *
 */
int		Requests	= DEF_REQUESTS;



/***************************************************************************
 *
 *
 *
 */

typedef struct {
	pthread_mutex_t	mutex1;
	pthread_mutex_t	mutex2;
	long		count1;
	long		count2;
	} Data_t;



/***************************************************************************
 *
 *
 *
 */

int shmemInit( const char * pShmemFile, int * pFd, Data_t ** ppData )
{
	/* open file */
	if( (*pFd = shm_open( pShmemFile, O_RDWR, S_IRUSR | S_IWUSR )) < 0 ) {
		return -1;
	}

	/* mmap */
	if( !(*ppData = (Data_t *) mmap( NULL, SHMEM_FILE_SIZE, PROT_WRITE | PROT_READ, MAP_SHARED, *pFd, 0 )) ) {
		close( *pFd );
		*pFd = -1;
		return -2;
	}

	return 0;
}

int shmemShutdown( int fd, void * pData )
{
	/* unmap */
	munmap( pData, SHMEM_FILE_SIZE );

	/* close */
	close( fd );

	return 0;
}



/***************************************************************************
 *
 *
 *
 */
int serverTask( const char * pShmemFile, int requests )
{
	int		request;
	int		fd;
	Data_t *	pData;
	long		val;


	/* init shmem handling */
	if( shmemInit( pShmemFile, &fd, &pData ) < 0 ) {
		/* failed */
		return 1;
	}

	/* initially lock */
#ifndef WITH_MSG
	pthread_mutex_trylock( &(pData->mutex1) );
#endif
	/* request loop */
	request = 0;
	while( request < requests ) {
		/* obtain mutex */
		if( pthread_mutex_lock( &(pData->mutex1) ) ) {
			shmemShutdown( fd, pData );
			return 2;
		}
#ifdef WITH_MSG
                if (msgsnd(msqid1, &sbuf, buf_length, 0 ) < 0) {
                       perror("msgsnd");
                  }
#endif
#ifndef WITH_MSG
		/* increment counter */
		pData->count1++;
#endif

		/* check counter */
		if( (val = pData->count1) != 0 ) {
			fprintf( stderr, "*** %05d: server: invalid counter value %d !\n", getpid(), val );
		}

		/* unlock mutex */
#ifdef WITH_MSG 
		/* increment counter */
		pData->count1++;
		pthread_mutex_unlock( &(pData->mutex1) );
#else
		pthread_mutex_unlock( &(pData->mutex2) );
#endif
		/* increment */
		request++;
#ifdef WITH_MSG
           if (msgrcv(msqid2, &rbuf, MSGSZ, 1, 0 ) < 0) {
                      perror("msgrcv");
             }
#endif
       } 
	/* shutdown shared memory handling */
	shmemShutdown( fd, pData );

	return 0;
}

int clientTask( const char * pShmemFile, int requests )
{
	int		request;
	int		fd;
	Data_t *	pData;
	long		val;
        long i,j = 0;
	/* init shmem handling */
	if( shmemInit( pShmemFile, &fd, &pData ) < 0 ) {
		/* failed */
		return 1;
	}
	/* request loop */
	request = 0;
	while( request < requests ) {
		/* obtain mutex */

#ifdef WITH_MSG
               if (msgrcv(msqid1, &rbuf, MSGSZ, 1, 0 ) < 0) {
                      perror("msgrcv");
                }
#endif 
#ifdef WITH_MSG
		if( pthread_mutex_lock( &(pData->mutex1) ) ) {
			shmemShutdown( fd, pData );
			return 2;
		}
		/* decrement counter */
		pData->count1--;
#else
		if( pthread_mutex_lock( &(pData->mutex2) ) ) {
			shmemShutdown( fd, pData );
			return 2;
		}
#endif 

		/* check counter */
		if( (val = pData->count1) != 0 ) {
			fprintf( stderr, "*** %05d: client: invalid counter value %d !\n", getpid(), val );
		}
#ifdef WITH_MSG
                if (msgsnd(msqid2, &sbuf, buf_length, 0 ) < 0) {
                       printf ("%d, %d, %s, %d\n", msqid2, sbuf.mtype, sbuf.mtext, 1);
                       perror("msgsnd");
                }

#endif
#ifndef WITH_MSG
		/* decrement counter */
		pData->count1--;
#endif
		/* release mutex */
		pthread_mutex_unlock( &(pData->mutex1) );

		/* increment */
		request++;
	}

	/* shutdown file watching */
	shmemShutdown( fd, pData );

	return 0;
}



/***************************************************************************
 *
 *
 *
 */
int setupMutexInSharedMem( void )
{
	int			fd;
	void *			pSharedArea;
	Data_t *		pData;
	pthread_mutexattr_t	mutexAttr;


	/* create file */
	if( (fd = shm_open( SHMEM_FILE_PATH, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR )) < 0 ) {
		fprintf( stderr, "*** creating file %s failed !\n", SHMEM_FILE_PATH );
		return 1;
	}
	
	/* set file size */
	if( ftruncate( fd, SHMEM_FILE_SIZE ) < 0 ) {
		fprintf( stderr, "*** setting file to %d bytes failed !\n", SHMEM_FILE_SIZE );
		close( fd );
		return 2;
	}

	/* mmap file */
	if( !(pSharedArea = mmap( NULL, SHMEM_FILE_SIZE, PROT_WRITE | PROT_READ, MAP_SHARED, fd, 0 )) ) {
		fprintf( stderr, "*** mapping file failed !\n" );
		close( fd );
		return 3;
	}
	pData = (Data_t *) pSharedArea;

	/* init mutex attribute */
	pthread_mutexattr_init( &mutexAttr );
	if( pthread_mutexattr_setpshared( &mutexAttr, PTHREAD_PROCESS_SHARED ) ) {
		fprintf( stderr, "*** setting mutex attribute to shared failed !\n" );
		close( fd );
		return 4;
	}

	/* init mutex 1 */
	memset( (void *) &(pData->mutex1), 0, sizeof(pthread_mutex_t) );
	if( pthread_mutex_init( &(pData->mutex1), &mutexAttr ) ) {
		fprintf( stderr, "*** init mutex failed !\n" );
		close( fd );
		return 5;
	}

	/* init mutex 2 */
	memset( (void *) &(pData->mutex2), 0, sizeof(pthread_mutex_t) );
	if( pthread_mutex_init( &(pData->mutex2), &mutexAttr ) ) {
		fprintf( stderr, "*** init mutex failed !\n" );
		close( fd );
		return 6;
	}

	/* init counter */
	pData->count1 = 0;
	pData->count2 = 0;

	/* munmap */
	munmap( pSharedArea, SHMEM_FILE_SIZE );

	/* close file */
	close( fd );

	return 0;
}



/***************************************************************************
 *
 *
 *
 */
void printUsage( char * cmd )
{
	printf( "%s: command line options\n", cmd );
	printf( "%s [-r <requests>]\n", cmd );
	printf( "\t-r <requests>    # of requests (%d..%d, default %d)\n",
		MIN_REQUESTS, MAX_REQUESTS, DEF_REQUESTS );

	return;
}

int parseCmdline( int argc, char * argv[] )
{
	int	option_r = 0;
	int	loop;
	int	state;

	state = 0;
	for( loop = 1; loop < argc; loop++ ) {
		switch( state ) {
			case 0:
				/* fetch option tag */
				if( !(strncmp( argv[loop], "-r", 3 )) ) {
					option_r = 1;
					state    = 1;
				} else {
					return -1;
				}
				break;

			case 1:
				/* fetch request count */
				sscanf( argv[loop], "%d", &Requests );
				state = 0;
				break;
		}
	}

	if( state ) {
		return -1;
	}

	if( (Requests < MIN_REQUESTS) || (Requests > MAX_REQUESTS) ) {
		return -1;
	}

	return 0;
}

int main( int argc, char * argv[] )
{
	pid_t		parent_pid;
	pid_t		pid;
	int		ret;
	int		status;
	struct timeval	start;
	struct timeval	stop;
	unsigned long	sec;
	long		usec;
	double		deltaTime;
	double		transactionsPerSec;


	/* parse command line */
	if( parseCmdline( argc, argv ) ) {
		printUsage( argv[0] );
		exit( 0 );
	}
	printf( "--- test configuration\n" );
	printf( "--- # requests : %d\n", Requests );

	/* setup mutex in shared memory */
	if( setupMutexInSharedMem() ) {
		fprintf( stderr, "*** preparing shared memory area for pthread mutex failed !\n" );
		exit( 1 );
	}
#ifdef WITH_MSG
        if ((msqid1 = msgget(KEY1,IPC_CREAT| 0666)) < 0) {
              perror("cmsgget");
        }
        if ((msqid2 = msgget(KEY2,IPC_CREAT| 0666)) < 0) {
              perror("cmsgget");
        }
        printf("msqid1 =%d msqid2 =%d\n",msqid1,msqid2);
        sbuf.mtype = 1;
 
       (void) strcpy(sbuf.mtext, "Did you get this?");
        buf_length = strlen(sbuf.mtext) + 1 ;
#endif

	/* spawn process */
	parent_pid = getpid();
	pid = fork();
	if( pid == 0 ) {
		/* child process */
		printf( "--- %05d: server process started\n", getpid() );
		fflush( stdout );
		ret = serverTask( SHMEM_FILE_PATH, Requests );
		if( ret )  {
			fprintf( stderr, "*** server failed !\n" );
		}
		printf( "--- %05d: server process stopped\n", getpid() );
		fflush( stdout );
		exit( ret );
	} else if (pid < 0) {
		/* failed */
		printf( "test failed: errno=%d\n", errno );
		exit( 0 );
	}

	/* give server some time to settle */
#ifndef WITH_MSG
	sleep( 1 );
#endif

	/* parent process */
	gettimeofday( &start, NULL );
	ret = clientTask( SHMEM_FILE_PATH, Requests );
	if( ret ) {
		fprintf( stderr, "*** client failed: %d (errno=%d) !\n", ret, errno );
	}
	gettimeofday( &stop, NULL );

	/* wait for termination of child process */
	waitpid( pid, &status, 0 );

	if( (status & 0xFF) ) {
		fprintf( stderr, "\t*** server stopped by signal %d\n", (status & 0xFF) );
		sleep( 1 );
	} else {
		status = (status >> 8) & 0xFF;
		if( status ) {
		fprintf( stderr, "\t*** server failed with return code 0x%02x\n", status );
		}
	}

	sec  = stop.tv_sec  - start.tv_sec;
	usec = stop.tv_usec - start.tv_usec;
	if( usec < 0 ) {
		usec = 1000000 + usec;
		sec--;
	}
	deltaTime = (double) usec;
	deltaTime = (deltaTime / 1000000.0) + ((double) sec);
	transactionsPerSec = ((double) Requests) / deltaTime;
	printf( "test done in %lu.%06lu seconds (%d requests, %lf requests/s)\n",
	        sec, usec, Requests, transactionsPerSec );

	exit( 0 );
}



#ifdef __cplusplus
}
#endif

